open Signatures

(* An implementation of identifiers as strings. [basename] chops off
   all trailing digits, while [combine] appends a representation of
   the integer salt grain to the basename. *)

module String = struct

  type t = string

  let compare = String.compare

  let basename s =
    let rec loop i =
      if i = 0 then
	assert false (* [s] is composed solely of digits. I guess one could return 0 here. *)
      else
	let im1 = i - 1 in
	match s.[im1] with
	| '0' .. '9' ->
	    loop im1
	| _ ->
	    i
    in
    let n = String.length s in
    let i = loop n in
    if i = n then
      s
    else
      String.sub s 0 i

  let combine s i =
    if i = 0 then
      s
    else
      s ^ (string_of_int i)

  module Map = Map.Make(String)

end

(* An implementation of atoms. *)

module Make (Identifier : Identifier) = struct

  type identifier = Identifier.t

  module Atom = struct

    (* An atom is implemented as a pair of an integer, the atom's
       identity, and an identifier, the atom's basename.

       We make identifiers part of atoms, instead of storing them in a
       separate table, so as to avoid the need for collecting garbage in
       such a table. *)

    type t = {
	identity: int;
	basename: Identifier.t
      }

    (* An integer counter holds the next available atom. *)

    let counter =
      ref 0

    let allocate () =
      let number = !counter in
      counter := number + 1;
      assert (number >= 0);
      number

    (* [freshb identifier] returns a fresh atom with basename [basename
       identifier]. *)

    let freshb identifier = {
      identity = allocate();
      basename = Identifier.basename identifier
    }

    (* [fresha a] returns a fresh atom whose basename is that of the
       atom [a]. *)

    let fresha a = {
      identity = allocate();
      basename = a.basename
    }

    (* Comparison over atoms. *)

    let equal a b =
      a.identity = b.identity

    let compare a b =
      a.identity - b.identity

    let hash a =
      Hashtbl.hash a.identity

    (* [identity a] maps every atom to a distinct integer. This number
       can be viewed as the atom's identity and can be turned into a
       unique name for printing purposes, if desired. *)

    let identity a =
      a.identity

    (* An atom can be queried about its base identifier. *)

    let basename a =
      a.basename

    exception Unknown of t

  end

  open Atom

  (* Efficient implementations of maps and sets over integers. *)

  module IntMap = Patricia.Big
  module IntSet = IntMap.Domain

  module AtomSet = struct

    (* Sets of atoms are represented as maps of integers to atoms. *)

    type t = Atom.t IntMap.t

    let empty =
      IntMap.empty

    let add a s =
      IntMap.add a.identity a s

    let singleton a =
      add a empty

    let union =
      IntMap.union

    let mem a s =
      try
        let _ = IntMap.lookup a.identity s in
	true
      with Not_found ->
	false

    let inter s1 s2 =
      IntMap.fold (fun identity a s1 ->
	try
	  let _ = IntMap.lookup identity s2 in
	  IntMap.add identity a s1
	with Not_found ->
	  s1
      ) s1 empty

    let is_empty =
      IntMap.is_empty

    let disjoint s1 s2 =
      IntMap.Domain.disjoint (IntMap.domain s1) (IntMap.domain s2)

    let equal s1 s2 =
      IntMap.Domain.equal (IntMap.domain s1) (IntMap.domain s2)

    let subset s1 s2 =
      IntMap.Domain.subset (IntMap.domain s1) (IntMap.domain s2)

    let fold f s accu =
      IntMap.fold (fun _ a accu ->
	f a accu
      ) s accu

    let iter f s =
      IntMap.iter (fun _ a ->
	f a
      ) s

    let diff s1 s2 =
      IntMap.corestrict s1 (IntMap.domain s2)

  end

  module AtomMap = struct

    (* Maps over atoms are represented as maps over integers. *)

    (* TEMPORARY il faudrait conserver le basename pour pouvoir
       implmenter domain, iter, fold, etc. *)
    (* TEMPORARY implmenter lift pour transformer un ensemble
       en un map; exige dans GMap un map qui donne accs aux clefs? *)
    (* TEMPORARY implmenter restrict/corestrict? *)

    type 'a t = 'a IntMap.t

    let empty =
      IntMap.empty

    let is_empty =
      IntMap.is_empty

    let singleton a v =
      IntMap.singleton a.identity v

    let add a v m =
      IntMap.add a.identity v m

    let union =
      IntMap.union

    let lookup a m =
      IntMap.lookup a.identity m

    let map =
      IntMap.map

  end

  module AtomIdMap = struct

    (* A map of atoms to identifiers is implemented as a map of atom
       identities (that is, integers) to identifiers, together with a
       map of basenames to so-called salt grains (integer counters). *)

    type t = {
	mapping: Identifier.t IntMap.t;
	salt: int Identifier.Map.t
      }

    (* The empty map. *)

    let empty = {
      mapping = IntMap.empty;
      salt = Identifier.Map.empty
    }

    (* The map [add a m] extends [m] by mapping [a] to a unique
       identifier. [a]'s basename is used in the computation of
       this identifier. *)

    let add a m =
      (* Check that [a] is not in the domain of [m]. *)
      assert (not (IntMap.Domain.mem a.identity (IntMap.domain m.mapping)));
      let basename = a.basename in
      let salt = m.salt in
      let grain =
	try
	  Identifier.Map.find basename salt
	with Not_found ->
	  0
      in {
	mapping = IntMap.add a.identity (Identifier.combine basename grain) m.mapping;
	salt = Identifier.Map.add basename (grain + 1) salt
      }

    (* [add_set s m] extends [m] with a mapping of [a] to a unique
       identifier for every atom [a] in the set [s]. *)

    let add_set s m =
      AtomSet.fold add s m

    (* [lookup a m] is the identifier associated with [a] in the
       map [m]. [Atom.Unknown] is raised if [a] is not a member of
       the domain of [m]. *)

    let lookup a m =
      try
	IntMap.lookup a.identity m.mapping
      with Not_found ->
	raise (Atom.Unknown a)

  end

  module Subst = struct

    type t = Atom.t AtomMap.t

    let id = AtomMap.empty
    let is_id = AtomMap.is_empty
    let singleton = AtomMap.singleton
    let add = AtomMap.add
    let union = AtomMap.union

    let lookup a subst =
      try
	AtomMap.lookup a subst
      with Not_found ->
	a

    let compose subst1 subst2 =

      (* If a mapping of [a] to [b] appears in [subst2], then a
	 mapping of [a] to [lookup b subst1] must appear in the
	 composition. This is the meaning of the second operand of
	 [union]. For atoms [a] that are not in the domain of
	 [subst2], we keep their image through [subst1]. This is the
	 meaning of the first operand. *)

      IntMap.union subst1 (IntMap.endo_map (fun b -> lookup b subst1) subst2)

    let freshen s subst =
      AtomSet.fold (fun a subst ->
	add a (Atom.fresha a) subst
      ) s subst

  end

end

